{% extends 'layout.html' %}
{% load i18n %}

{% block title %}{% trans 'Password reset successful' %} | {{ block.super }}{% endblock %}

{% block content %}

    <h1>{% trans 'Password reset successful' %}</h1>
    <p>{% trans "We've e-mailed you instructions for setting your password to the e-mail address you submitted. You should be receiving it shortly." %}</p>

{% endblock %}
