{% extends 'layout.html' %}
{% load i18n %}

{% block title %}{% trans 'Password reset complete' %} | {{ block.super }}{% endblock %}

{% block content %}
	<h1>{% trans 'Password reset complete' %}</h1>
	<p>{% trans "Your password has been set.  You may go ahead and log in now." %}</p>
	<p><a href="{{ login_url }}">{% trans 'Log in' %}</a></p>
{% endblock %}
